# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import ast
import logging
import os

import cv2
import numpy as np

from tqdm import tqdm

logger = logging.getLogger(__name__)


def _compute_max_iou_and_ioda(detected_box, gt_boxes):
    """Will calculate how much of detected area lies in ground truth boxes.

    If there are multiple gt boxes, the detected area will be calculated
    for each gt box separately and the gt box with the highest
    intersecting part will be used for the calculation.
    """
    x_L, y_L, w_L, h_L = detected_box
    detected_area = w_L * h_L
    if detected_area == 0:
        return 0, 0
    max_intersection = 0
    max_gt_area = 0

    for bbox in gt_boxes:
        unpacked_bbox = bbox[0]
        bbox_dict = ast.literal_eval(unpacked_bbox.replace("'", '"'))
        xmin, ymin = int(bbox_dict["xmin"]), int(bbox_dict["ymin"])
        width, height = int(bbox_dict["width"]), int(bbox_dict["height"])

        gt_area = width * height

        # Calculate intersection
        xi = max(x_L, xmin)
        yi = max(y_L, ymin)
        wi = max(0, min(x_L + w_L, xmin + width) - xi)
        hi = max(0, min(y_L + h_L, ymin + height) - yi)

        intersection = 0
        if wi > 0 and hi > 0:
            intersection = wi * hi

        if intersection > max_intersection:
            max_intersection = intersection
            max_gt_area = gt_area

    if max_gt_area == 0 and max_intersection == 0:
        # This case means no intersection was found, even though there are gt boxes
        iou, ioda = 0, 0
    else:
        iou = max_intersection / (
            detected_area + max_gt_area - max_intersection
        )
        ioda = max_intersection / detected_area

    return iou, ioda


def _compute_simultaneous_iou_and_ioda(detected_box, gt_boxes):
    """Will calculate how much of detected area lies between ground truth
    boxes.

    This means that if there are multiple gt boxes, the detected area
    will be compared to them simultaneously (and not to each gt box
    separately).)
    """
    x_L, y_L, w_L, h_L = detected_box
    detected_area = w_L * h_L
    if detected_area == 0:
        return 0, 0
    intersection = 0
    total_gt_area = 0

    for bbox in gt_boxes:
        unpacked_bbox = bbox[0]
        bbox_dict = ast.literal_eval(unpacked_bbox.replace("'", '"'))
        xmin, ymin = int(bbox_dict["xmin"]), int(bbox_dict["ymin"])
        width, height = int(bbox_dict["width"]), int(bbox_dict["height"])

        gt_area = width * height
        total_gt_area += gt_area

        # Calculate intersection
        xi = max(x_L, xmin)
        yi = max(y_L, ymin)
        wi = max(0, min(x_L + w_L, xmin + width) - xi)
        hi = max(0, min(y_L + h_L, ymin + height) - yi)

        if wi > 0 and hi > 0:
            intersection += wi * hi

    iou = intersection / (detected_area + total_gt_area - intersection)
    ioda = intersection / detected_area

    return iou, ioda


def _compute_avg_saliency_focus(gt_boxes, saliency_map):
    """Will calculate how much of the ground truth bounding boxes area is
    covered by the activations."""

    binary_mask = np.zeros_like(saliency_map)

    total_gt_bbox_area = 0

    # For each gt box, draw a binary mask
    # The binary_mask will be 1 where the gt boxes are located
    for bbox in gt_boxes:
        unpacked_bbox = bbox[0]
        bbox_dict = ast.literal_eval(unpacked_bbox.replace("'", '"'))
        xmin, ymin = int(bbox_dict["xmin"]), int(bbox_dict["ymin"])
        width, height = int(bbox_dict["width"]), int(bbox_dict["height"])

        binary_mask[ymin : ymin + height, xmin : xmin + width] = 1

        total_gt_bbox_area += width * height

    multiplied_mask = binary_mask * saliency_map
    numerator = np.sum(multiplied_mask)

    if total_gt_bbox_area == 0:
        avg_saliency_focus = 0
    else:
        avg_saliency_focus = numerator / total_gt_bbox_area

    return avg_saliency_focus


# Own implementation based on
# "Score-CAM: Score-Weighted Visual Explanations for Convolutional Neural Networks" by Wang et al. (2020),
# https://arxiv.org/abs/1910.01279
def _compute_proportional_energy(gt_boxes, saliency_map):
    """Will calculate how much activation lies within the ground truth boxes
    compared to the total sum of the activations."""
    binary_mask = np.zeros_like(saliency_map)

    # For each gt box, draw a binary mask
    # The binary_mask will be 1 where the gt boxes are located
    for bbox in gt_boxes:
        unpacked_bbox = bbox[0]
        bbox_dict = ast.literal_eval(unpacked_bbox.replace("'", '"'))
        xmin, ymin = int(bbox_dict["xmin"]), int(bbox_dict["ymin"])
        width, height = int(bbox_dict["width"]), int(bbox_dict["height"])

        binary_mask[ymin : ymin + height, xmin : xmin + width] = 1

    multiplied_mask = binary_mask * saliency_map
    numerator = np.sum(multiplied_mask)
    denominator = np.sum(saliency_map)

    if denominator == 0:
        proportional_energy = 0
    else:
        proportional_energy = numerator / denominator

    return proportional_energy


def calculate_localization_metrics(
    saliency_map, detected_box, ground_truth_box
):
    """Calculates localization metrics for a single input image for a given
    visualization method."""

    iou, ioda = _compute_max_iou_and_ioda(detected_box, ground_truth_box)

    proportional_energy = _compute_proportional_energy(
        ground_truth_box, saliency_map
    )

    avg_saliency_focus = _compute_avg_saliency_focus(
        ground_truth_box, saliency_map
    )

    return iou, ioda, proportional_energy, avg_saliency_focus


# Helper function to calculate the metrics for a single target class
# of a single input image.
def process_target_class(
    names,
    gt_bboxes,
    saliency_map_path,
    csv_writer,
):
    saliency_map = np.load(saliency_map_path)

    # Calculate bounding boxes for largest connected component
    # The pixel values above 20% of max value are kept in the mask to
    # calculate IoU and IoDA.
    # This imitates the process done by the original CAM paper:
    # "Learning Deep Features for Discriminative Localization" by Zhou et al. (2015),
    # https://arxiv.org/abs/1512.04150
    thresholded_mask = np.copy(saliency_map)
    max_value = np.max(thresholded_mask)
    threshold_value = 0.2 * max_value
    thresholded_mask[thresholded_mask < threshold_value] = 0
    thresholded_mask = (thresholded_mask > 0).astype(np.uint8)
    if np.any(thresholded_mask > 0):
        _, label_ids, values, _ = cv2.connectedComponentsWithStats(
            thresholded_mask, connectivity=8
        )
        largest_label_id = np.argmax(values[1:, cv2.CC_STAT_AREA]) + 1
        largest_label_mask = (label_ids == largest_label_id).astype(np.uint8)
        x_L, y_L, w_L, h_L = cv2.boundingRect(largest_label_mask)
        x_L, y_L, w_L, h_L = int(x_L), int(y_L), int(w_L), int(h_L)
    else:
        x_L, y_L, w_L, h_L = 0, 0, 0, 0

    # Calculate localization metrics
    iou, ioda, proportional_energy, asf = calculate_localization_metrics(
        saliency_map=saliency_map,
        detected_box=(x_L, y_L, w_L, h_L),
        ground_truth_box=gt_bboxes,
    )

    # Write metrics to csv file
    csv_writer.writerow(
        [
            names[0].split("/")[1],
            iou,
            ioda,
            proportional_energy,
            asf,
            x_L,
            y_L,
            w_L,
            h_L,
        ]
    )


def run(
    input_folder,
    data_loader,
    dataset_split_name,
    csv_writers,
):
    """Applies visualization techniques on input CXR, outputs images with
    overlaid heatmaps and csv files with measurements.

    Parameters
    ---------

    input_folder : str
        Directory in which the saliency maps are stored for a specific visualization type.

    data_loader : py:class:`torch.torch.utils.data.DataLoader`
        The pytorch Dataloader used to iterate over batches.

    dataset_split_name : str
        Name of the dataset split (e.g. "train", "validation", "test").

    csv_writers : dict
        Dictionary containing csv writer objects for each target class.

    Returns
    -------

    all_predictions : list
        All the predictions associated with filename and ground truth, saved as .csv.
    """
    for samples in tqdm(data_loader, desc="batches", leave=False, disable=None):
        # Check if the sample has a bounding box entry
        if len(samples) < 4:
            logger.warning(
                "The dataset does not contain bounding box information. No localization metrics can be calculated."
            )
            return
        else:
            # TB negative labels are skipped (they don't have gt bboxes)
            if samples[3][0] == "none":
                continue

        names = samples[0]

        gt_bboxes = samples[3]

        for target_class_name, csv_writer in csv_writers.items():
            saliency_map_path = os.path.join(
                input_folder,
                target_class_name,
                dataset_split_name,
                names[0].rsplit(".", 1)[0] + ".npy",
            )

            process_target_class(
                names,
                gt_bboxes,
                saliency_map_path,
                csv_writer,
            )
