# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import os

import numpy as np
import torch

from pytorch_grad_cam.utils.model_targets import ClassifierOutputTarget
from tqdm import tqdm

logger = logging.getLogger(__name__)

rs_maps = {
    0: "cardiomegaly",
    1: "emphysema",
    2: "effusion",
    3: "hernia",
    4: "infiltration",
    5: "mass",
    6: "nodule",
    7: "atelectasis",
    8: "pneumothorax",
    9: "pleural thickening",
    10: "pneumonia",
    11: "fibrosis",
    12: "edema",
    13: "consolidation",
}


def _save_npy(img_name_stem, grayscale_cam, visualization_path):
    image_path = f"{visualization_path}/{img_name_stem}"
    os.makedirs(os.path.dirname(image_path), exist_ok=True)

    np.save(image_path, grayscale_cam)


# Helper function to calculate saliency maps for a single target class
# of a single input image.
def process_target_class(names, images, targets, visualization_path, cam):
    grayscale_cams = cam(input_tensor=images, targets=targets)

    for i, grayscale_cam in enumerate(grayscale_cams):
        img_name_stem = names[i].split(".")[0]

        _save_npy(img_name_stem, grayscale_cam, visualization_path)


def run(
    model,
    data_loader,
    output_folder,
    dataset_split_name,
    device,
    cam,
    visualization_type,
    target_class="highest",
    tb_positive_only=True,
):
    """Applies visualization techniques on input CXR, outputs pickled saliency
    maps.

    Parameters
    ---------
    model : py:class:`torch.nn.Module`
        The model to use for the saliency map calculation.

    data_loader : py:class:`torch.torch.utils.data.DataLoader`
        The pytorch Dataloader used to iterate over batches.

    output_folder : str
        Directory in which the results will be saved.

    dataset_split_name : str
        Name of the dataset split (e.g. "train", "validation", "test").

    device : str
        A string indicating the device to use (e.g. "cpu" or "cuda"). The device can also be specified (cuda:0)

    cam : py:class: `pytorch_grad_cam.GradCAM`, `pytorch_grad_cam.ScoreCAM`,
    `pytorch_grad_cam.FullGrad`, `pytorch_grad_cam.RandomCAM`,
    `pytorch_grad_cam.EigenCAM`, `pytorch_grad_cam.EigenGradCAM`,
    `pytorch_grad_cam.LayerCAM`, `pytorch_grad_cam.XGradCAM`,
    `pytorch_grad_cam.AblationCAM`, `pytorch_grad_cam.HiResCAM`,
    `pytorch_grad_cam.GradCAMElementWise`, `pytorch_grad_cam.GradCAMplusplus`,
        The CAM object to use for visualization.

    visualization_types : list
        Type of visualization techniques to be applied. Possible values are:
        "GradCAM", "ScoreCAM", "FullGrad", "RandomCAM", "HiResCAM", "GradCAMElementWise", "GradCAMPlusPlus", "XGradCAM", "AblationCAM",
        "EigenCAM", "EigenGradCAM", "LayerCAM".

    target_class : str
        (Use only with multi-label models) Which class to target for CAM calculation. Can be either set to "all" or "highest". "highest" is default, which means only saliency maps for the class with the highest activation will be generated.

    tb_positive_only : bool
        If set, saliency maps will only be generated for TB positive samples.

    Returns
    -------

    all_predictions : list
        All the predictions associated with filename and ground truth, saved as .csv.
    """
    output_folder = os.path.abspath(output_folder)

    logger.info(f"Output folder: {output_folder}")
    os.makedirs(output_folder, exist_ok=True)

    model_name = model.__class__.__name__

    for samples in tqdm(data_loader, desc="batches", leave=False, disable=None):
        # TB negative labels are skipped (they don't have bboxes)
        if samples[2][0].item() == 0:
            if tb_positive_only:
                continue

        names = samples[0]
        images = samples[1].to(
            device=device, non_blocking=torch.cuda.is_available()
        )

        if model_name == "DensenetRS" and target_class.lower() == "all":
            for target in range(14):
                targets = [ClassifierOutputTarget(target)]

                visualization_path = f"{output_folder}/{visualization_type}/{rs_maps[target]}/{dataset_split_name}"
                os.makedirs(visualization_path, exist_ok=True)

                process_target_class(
                    names, images, targets, visualization_path, cam
                )

        if model_name == "DensenetRS":
            # Get the class with highest activation manually
            outputs = cam.activations_and_grads(images)
            target_categories = np.argmax(outputs.cpu().data.numpy(), axis=-1)
            targets = [
                ClassifierOutputTarget(category)
                for category in target_categories
            ]
        else:
            targets = [ClassifierOutputTarget(0)]

        visualization_path = f"{output_folder}/{visualization_type}/targeted_class/{dataset_split_name}"
        os.makedirs(visualization_path, exist_ok=True)

        process_target_class(names, images, targets, visualization_path, cam)
