# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import os

import cv2
import numpy as np
import pandas as pd

from PIL import Image
from tqdm import tqdm

from ..utils.cam_utils import (
    draw_boxes_on_image,
    draw_largest_component_bbox_on_image,
    show_cam_on_image,
    visualize_road_scores,
)

logger = logging.getLogger(__name__)


def _get_target_classes_from_directory(input_folder):
    # Gets a list of target classes from a directory
    return [
        item
        for item in os.listdir(input_folder)
        if os.path.isdir(os.path.join(input_folder, item))
    ]


def run(
    data_loader,
    img_dir,
    input_folder,
    output_folder,
    dataset_split_name,
    visualize_groundtruth=False,
    road_path=None,
    visualize_detected_bbox=True,
    threshold=0.0,
):
    """Overlays saliency maps on CXR to output final images with heatmaps.

    Parameters
    ---------
    img_dir : str
        Directory containing the images from the dataset to use for visualization.

    input_folder : str
        Directory in which the saliency maps are stored for a specific visualization type.

    output_folder : str
        Directory in which the results will be saved.

    visualize_groundtruth : bool
        If set, generate visualizations for ground truth labels.

    road_path : str
        If the path to the previously calculated road scores is provided, MoRF, LeRF, and combined ROAD scores will be visualized on each image.

    visualize_detected_bbox : bool
        If set, the bounding box for the largest connected component will be visualized on each image.

    threshold : float
        Only CAMs above this threshold will be visualized.

    Returns
    -------

    output_folder : str
        Directory containing the images overlaid with heatmaps.
    """
    assert (
        input_folder != output_folder
    ), "Output folder must not be the same as the input folder."

    # Check that the output folder is not a subdirectory of the input_folder
    assert not output_folder.startswith(
        input_folder
    ), "Output folder must not be a subdirectory of the input folder."

    output_folder = os.path.abspath(output_folder)

    logger.info(f"Output folder: {output_folder}")
    os.makedirs(output_folder, exist_ok=True)

    target_classes = _get_target_classes_from_directory(input_folder)

    if visualize_detected_bbox:
        # Load detected bounding box information
        localization_metric_dfs = {}

        for target_class in target_classes:
            input_base_path = f"{input_folder}/{target_class}/{dataset_split_name}_localization_metrics.csv"

            if os.path.exists(input_base_path):
                df = pd.read_csv(input_base_path)
                localization_metric_dfs[target_class] = df

    if road_path is not None:
        # Load road score information
        road_metric_dfs = {}

        for target_class in target_classes:
            road_base_path = f"{road_path}/{target_class}/{dataset_split_name}_road_metrics.csv"

            if os.path.exists(road_base_path):
                df = pd.read_csv(road_base_path)
                road_metric_dfs[target_class] = df

    for samples in tqdm(data_loader, desc="batches", leave=False, disable=None):
        includes_bboxes = True
        if len(samples) < 4 or samples[2][0].item() == 0:
            includes_bboxes = False

        # if visualize_groundtruth and not includes_bboxes:
        #    logger.warning(
        #        "This sample does not have bounding box information. No ground truth bounding boxes can be visualized."
        #    )

        names = samples[0]
        gt_bboxes = samples[3]

        for target_class in target_classes:
            input_base_path = (
                f"{input_folder}/{target_class}/{dataset_split_name}"
            )
            vis_directory_name = os.path.basename(input_folder)
            output_base_path = f"{output_folder}/{vis_directory_name}/{target_class}/{dataset_split_name}"

            img_path = os.path.join(img_dir, names[0])
            saliency_map_name = names[0].rsplit(".", 1)[0] + ".npy"
            saliency_map_path = os.path.join(input_base_path, saliency_map_name)

            # Skip if the saliency map does not exist (e.g. if no saliency maps were generated for TB negative samples)
            if not includes_bboxes and not os.path.exists(saliency_map_path):
                continue

            saliency_map = np.load(saliency_map_path)

            # Image is expected to be of type float32
            original_img = np.array(Image.open(img_path), dtype=np.float32)
            img_with_boxes = original_img.copy()

            # Draw bounding boxes on the image
            if includes_bboxes:
                if gt_bboxes != "none" and visualize_groundtruth:
                    img_with_boxes = draw_boxes_on_image(
                        img_with_boxes, gt_bboxes
                    )

            # show_cam_on_image expects the img to be between [0, 1]
            rgb_img = img_with_boxes / 255.0

            # Threshold heatmap
            thresholded_mask = np.copy(saliency_map)
            thresholded_mask[thresholded_mask < threshold] = 0

            # Visualize heatmap
            visualization = show_cam_on_image(
                rgb_img, thresholded_mask, use_rgb=False
            )

            # Visualize bounding box for largest connected component
            if visualize_detected_bbox:
                if target_class in localization_metric_dfs:
                    temp_df = localization_metric_dfs[target_class]
                    temp_filename = names[0].split("/")[1]

                    row = temp_df[temp_df["Image"] == temp_filename]

                    if not row.empty:
                        x_L = int(row["detected_bbox_xmin"].values[0])
                        y_L = int(row["detected_bbox_ymin"].values[0])
                        w_L = int(row["detected_bbox_width"].values[0])
                        h_L = int(row["detected_bbox_height"].values[0])

                        visualization = draw_largest_component_bbox_on_image(
                            visualization, x_L, y_L, w_L, h_L
                        )
                    else:
                        logger.warning(
                            f"Could not find entry for {temp_filename} in .csv file."
                        )
                else:
                    logger.warning(
                        f"No localization metrics .csv file found for the {dataset_split_name} split for {target_class}."
                    )

            # Visualize ROAD scores
            if road_path is not None:
                if target_class in road_metric_dfs:
                    temp_df = road_metric_dfs[target_class]
                    temp_filename = names[0].split("/")[1]

                    row = temp_df[temp_df["Image"] == temp_filename]

                    if not row.empty:
                        MoRF_score = float(row["MoRF"].values[0])
                        LeRF_score = float(row["LeRF"].values[0])
                        combined_score = float(
                            row["Combined Score ((LeRF-MoRF) / 2)"].values[0]
                        )
                        percentiles = str(row["Percentiles"].values[0])
                    else:
                        logger.warning(
                            f"Could not find entry for {temp_filename} in .csv file."
                        )
                else:
                    logger.warning(
                        f"No ROAD metrics .csv file found for the {dataset_split_name} split for {target_class}."
                    )

                visualization = visualize_road_scores(
                    visualization,
                    MoRF_score,
                    LeRF_score,
                    combined_score,
                    vis_directory_name,
                    percentiles,
                )

            # Save image
            output_file_path = os.path.join(output_base_path, names[0])
            os.makedirs(os.path.dirname(output_file_path), exist_ok=True)
            cv2.imwrite(output_file_path, visualization)
